__author__ = 'Martin Humby 2015'

from graphs import Graph, Vertex, Edge
from queues import Queue, MinPQueue

INF = 100

def dijkstra(graph, start):
    pq = MinPQueue()
    graph.resetDistances(INF)
    path = []                           # the path through graph
    start = graph.vmap[start]           # get mapped start vertex
    start.distance = 0
    pq.joinItems(graph.vertices)
    print('           initial priority queue', pq)
    while not pq.empty():
        vertex = pq.leave()
        print('remove%9s' % str(vertex), end=': updates queue to ')
        path.append((vertex.name, vertex.distance))
        for edge in vertex.edges:
            toVertex = edge.toVertex
            weight = edge.weight
            new_distance = vertex.distance + weight
            if new_distance < toVertex.distance:
                toVertex.distance = new_distance
                pq.update()
        print(pq)

    return path

def primsAlogorithm(graph, start):
    pq = MinPQueue()
    graph.resetDistances(INF)
    MST = []                        # minimum spanning tree
    weightSum = 0.0                 # the sum of edge weights added to the MST
    start = graph.vmap[start]       # get mapped start vertex
    start.distance = 0
    pq.joinItems(graph.vertices)
    print('           initial priority queue', pq)
    while not pq.empty():
        vertex = pq.leave()
        print('remove%9s' % str(vertex), end=': updates queue to ')

        # check all edges from the vertex to see if it is closer toVertex than a previous vertex
        for edge in vertex.edges:
            toVertex = edge.toVertex
            edge_weight = edge.weight
            # if it is closer and toVertex is still in the queue i.e. it is not already in the MST then ...
            if edge_weight < toVertex.distance and pq.includes(toVertex):
                toVertex.currentNearest = vertex
                toVertex.distance = edge_weight  # in Prim's, vertex.distance is vertex -> toVertex
            pq.update()
        print(pq)

        # the vertex at the head of the priority queue will now have the shortest available
        # distance from its currentNearest so add this shortest edge to the MST
        if not pq.empty():
            nextVertex = pq.front()
            newEdge = [[(nextVertex.currentNearest.name, nextVertex.name), nextVertex.distance]]
            MST += newEdge
            weightSum += nextVertex.distance
            print('               MST is now', MST)

    return MST, weightSum


if __name__ == '__main__':

    print('\nFor the graph from Unit 5 Figure 5.20:')

    g = Graph(directed = False)
    g.addEdges('w', ('u',2), ('z',4), ('x',2))
    g.addEdges('x', ('v',6), ('y',3))
    g.addEdges('u', ('v',1))
    g.addEdges('v', ('y',5))
    g.addEdges('y', ('z',1))
    g.addEdges('z', ('v',6))

    print('adjacencies')
    print(g.adjacencies())

    start = 'w'
    print('\nDijkstra: shortest distance from ' + str(start) + ' to vertices')
    print('\n',dijkstra(g, start))

    print("\nMST from Prim's algorithm")
    MST, weightsTotal = primsAlogorithm(g, start)
    print('\nMST:', MST)
    print('total weight of edges', weightsTotal)
    print()

    print("\nFor the graph from Katrine's tutorial 4/3/15 slide 52:")

    g = Graph(directed = False)
    g.addEdges('A', ('B',9), ('E',8), ('F',14))
    g.addEdges('B', ('E',8), ('F',7), ('C',14))
    g.addEdges('C', ('F',8), ('D',15))
    g.addEdges('D', ('F',12), ('E',12))
    g.addEdges('E', ('F',10))

    print('adjacencies')
    print(g.adjacencies())

    start = 'A'
    print('\nDijkstra: shortest distance from ' + str(start) + ' to vertices')
    print('\n',dijkstra(g, start))
    print()
    print("MST from Prim's algorithm")
    MST, weightsTotal = primsAlogorithm(g, start)
    print('\nMST:', MST)
    print('total weight of edges', weightsTotal)

    print("\n\nFor Graph 1 from Tyler A. Simon -- Adjunct Professor UBMC:")
    # http://www.csee.umbc.edu/~tsimo1/distalg/graphalg/graphalg.html
    g = Graph(directed = False)
    g.addEdges(0, (1,2),(2,4),(3,2),(4,8))
    g.addEdges(1, (2,2),(3,6),(4,6))
    g.addEdges(2, (3,9),(4,6))
    g.addEdges(3, (4,2))
    g.addEdges(4)

    print('adjacencies')
    print(g.adjacencies(),'\n')

    start = 0
    print("MST from Prim's algorithm")
    MST, weightsTotal = primsAlogorithm(g, start)
    print('\nMST:', MST)
    print('total weight of edges', weightsTotal)

    # if edge_weight < toVertex.distance and not pq.includes(toVertex):
    #     print('>',str(toVertex), end = ' ')