__author__ = 'Martin Humby 2015'

from graphs import Graph, Vertex, Edge
from queues import O1Queue

class TopEdge(Edge):
    def __init__(self, toVertex, weight = 0):
        Edge.__init__(self, toVertex, weight)
        self.toVertex.incoming += 1

class topVertex(Vertex):
    def __init__(self, name, ID):
        Vertex.__init__(self, name, ID)
        self.incoming = 0

    def nilIncoming(self):
        return self.incoming == 0

    def decIncoming(self):
        if self.incoming:
            self.incoming -= 1
            return True
        return False

def listVertices(graph):
    for v in graph:
        print((v.name, v.incoming), end='')
    print()

def addNilIncoming(to_queue, from_graph):
    for v in from_graph:
        if v.nilIncoming():
            to_queue.join(v)

def topSort(graph):
    listVertices(graph)             # print a list of vertices and their initial incoming edge counts
    incomingFree = O1Queue.Queue()  # vertices with no incoming edges
    topsort = []                    # the list of vertices in topological order

    addNilIncoming(incomingFree, graph)
    # while there are vertices in the queue of incoming free vertices
    while incomingFree:
        vertex = incomingFree.leave()
        topsort.append(vertex.name)
        # conceptually remove vertices from vertices reachable from the current vertex
        for edge in vertex.edges:
            tv = edge.toVertex
            if tv.decIncoming() and tv.nilIncoming():
                incomingFree.join(tv)

    return topsort

# when an edge  running to a vertex is added to the graph...
def incIncoming(fromVertex, edge):               # (fromVertex is not used in this context)
    edge.toVertex.incoming += 1



print('topSort the graph from Unit 5 Figure 5.27')
g = Graph(directed = True, vertexClass=topVertex, edgeClass = TopEdge)
g.addEdges('1', '2')
g.addEdges('2')
g.addEdges('3', '1','6')
g.addEdges('4', '3','5')
g.addEdges('5')
g.addEdges('6', '2', '5')

print(topSort(g), '\n')


print('topSort the graph from iCMA 45 Question 7')
g = Graph(directed = True, vertexClass=topVertex, edgeClass = TopEdge)

g.addEdges('TA', 'TB')
g.addEdges('TB', 'TC','TF')
g.addEdges('TC', 'TD')
g.addEdges('TD', 'TE')
g.addEdges('TE', 'TH')
g.addEdges('TF', 'TH')
g.addEdges('TG', 'TB','TH')

print(topSort(g), '\n')

g = Graph(directed = True, vertexClass=topVertex, edgeClass = TopEdge)

g.addEdges('TG', 'TB','TH')
g.addEdges('TA', 'TB')
g.addEdges('TB', 'TC','TF')
g.addEdges('TC', 'TD')
g.addEdges('TD', 'TE')
g.addEdges('TE', 'TH')
g.addEdges('TF', 'TH')

print(topSort(g))