__author__ = 'Martin Humby 2015'

from graphs import Graph, Vertex, Edge
from queues import O1Queue

def listVertices(graph):
    for v in graph:
        print((v.name, v.incoming), end='')
    print()

def decIncoming(v):
    if v.incoming:
        v.incoming -= 1
        return True
    return False

def topSort(graph):
    incomingFree = O1Queue.Queue()  # vertices with no incoming edges
    topsort = []                    # the list of vertices in topological order

    for v in graph:
        v.incoming = v.in_order()
        if not v.incoming:
            incomingFree.join(v)

    listVertices(graph)             # print a list of vertices and their initial incoming edge counts
    # while there are vertices in the queue of incoming free vertices
    while incomingFree:
        vertex = incomingFree.leave()
        topsort.append(vertex.name)
        # conceptually remove vertices from vertices reachable from the current vertex
        for edge in vertex.edges:
            tv = edge.toVertex
            if decIncoming(tv) and not tv.incoming:
                incomingFree.join(tv)

    for v in graph:
        del(v.incoming)

    return topsort



print('topSort the graph from Unit 5 Figure 5.27')
g = Graph(directed = True)
g.addEdges('1', '2')
g.addEdges('2')
g.addEdges('3', '1','6')
g.addEdges('4', '3','5')
g.addEdges('5')
g.addEdges('6', '2')
g.addEdges('6', '5')

print(topSort(g), '\n')



print('topSort the graph from iCMA 45 Question 7')
g = Graph(directed = True)

g.addEdges('TA', 'TB')
g.addEdges('TB', 'TC','TF')
g.addEdges('TC', 'TD')
g.addEdges('TD', 'TE')
g.addEdges('TE', 'TH')
g.addEdges('TF', 'TH')
g.addEdges('TG', 'TB','TH')

print(topSort(g), '\n')

g = Graph(directed = True)

g.addEdges('TG', 'TB','TH')
g.addEdges('TA', 'TB')
g.addEdges('TB', 'TC','TF')
g.addEdges('TC', 'TD')
g.addEdges('TD', 'TE')
g.addEdges('TE', 'TH')
g.addEdges('TF', 'TH')

print(topSort(g))