__author__ = 'Martin Humby 2015'

from graphs import Graph
from queues import Queue, MinPQueue

I32INF = 2147483647

def depthFirst(graph, start):
    start = graph.vmap[start]           # get mapped start vertex
    visited = [False] * graph.vcnt      # create lookup of nodes visited
    path = []                           # the path through graph
    _dfs(graph, start, visited, path)   # call recursive depth first
    return path

def _dfs(graph, vertex, visited, path):
    if not visited[vertex.ID]:
        visited[vertex.ID] = True
        path.append(vertex.name)
        # vertex.edges = sorted(vertex.edges)
    for v in vertex.edges:
        if not visited[v.toVertex.ID]:
            _dfs(graph, v.toVertex, visited, path)

def breadthFirst(graph, start):

    start = graph.vmap[start]           # get mapped start vertex
    visited = [False] * graph.vcnt      # create lookup of nodes visited
    path = []                           # the path through graph
    q = Queue()
    visited[start.ID] = True
    q.join(start)
    while not q.empty():
        vertex = q.leave()
        path.append(vertex.name)
        vertex.edges = sorted(vertex.edges)
        for v in vertex.edges:
            if not visited[v.toVertex.ID]:
                visited[v.toVertex.ID] = True
                q.join(v.toVertex)
    return path


def dijkstra(graph, start):
    pq = MinPQueue()
    graph.resetDistances(I32INF)
    path = []                           # the path through graph
    start = graph.vmap[start]           # get mapped start vertex
    start.distance = 0
    pq.joinItems(graph.vertices)
    while not pq.empty():
        u = pq.leave()
        path.append((u.name, u.distance))
        for edge in u.edges:
            w = edge.toVertex
            weight = edge.weight
            new_distance = u.distance + weight
            if new_distance < w.distance:
                w.distance = new_distance
        pq.update()

    return path


if __name__ == '__main__':

    print('\nFor the graph from Unit 5 Figure 5.20:')
    g = Graph(directed = False)
    g.addEdges('w', ('u',2), ('z',4), ('x',2))
    g.addEdges('x', ('v',6), ('y',3))
    g.addEdges('u', ('v',1))
    g.addEdges('y', ('v',5))
    g.addEdges('v', ('y',5))
    g.addEdges('z', ('y',1), ('v',6))

    start = 'w'
    print('depth-first: vertices visited')
    print(depthFirst(g, start))

    print('breadth-first: vertices visited')
    print(breadthFirst(g, start))

    print('Dijkstra: vertices shortest distance from ' + str(start))
    print(dijkstra(g, start))



    g = Graph(directed = False)
    g.addEdges('w', ('x',2), ('z',4))
    g.addEdges('x', ('v',6), ('y',3))
    g.addEdges('u', ('v',1))
    g.addEdges('y', ('v',5))
    g.addEdges('v', ('y',5))
    g.addEdges('z', ('y',1), ('v',6))
    g.addEdges('w', ('u',2))

    start = 'w'
    print('depth-first: vertices visited')
    print(depthFirst(g, start))

    print('breadth-first: vertices visited')
    print(breadthFirst(g, start))

    print('Dijkstra: vertices shortest distance from ' + str(start))
    print(dijkstra(g, start))

    print('\nFor the graph from Wikipedia')
    # http://en.wikipedia.org/wiki/Breadth-first_search#Pseudocode
    d = Graph(directed = False)
    d.addEdges('Fra', ('Man',85), ('Wur',217), ('Kas',173))
    d.addEdges('Man', ('Kar',80))
    d.addEdges('Kar', ('Aug',250))
    d.addEdges('Aug', ('Mun',84))
    d.addEdges('Wur', ('Erf',186), ('Nur',103))
    d.addEdges('Nur', ('Stu',183), ('Mun',167))
    d.addEdges('Kas', ('Mun',502))
    start = 'Fra'
    print('adjacencies')
    print(d.adjacencies())

    print('depth-first: vertices visited')
    print(depthFirst(d, start))
    print('Dijkstra: shortest distances from Frankfurt')
    print(dijkstra(d, start))

    print('\nFor the graph from iCMA 45, Question 3')
    g = Graph(directed = True)
    g.addEdges('E',  'A', 'B', 'L')
    g.addEdges('B', 'T', 'F')
    g.addEdges('E', 'B', 'L')
    g.addEdges('A', 'R', 'M')
    g.addEdges('F', 'M')
    g.addEdges('L', 'R')

    start = 'E'
    print('depth-first: vertices visited')
    print(depthFirst(g, start))
