__author__ = 'Martin Humby 2015'

"""
Queues with O(1) join and leave time complexity (nominal in the case of the priority queue;))
"""

class Queue:

    def __init__(self):
        self.back = []
        self.front = []

    def join(self, item):
        self.back.append(item)

    def leave(self):
        # pre: not empty
        if not self.front:
            self.back, self.front = self.front, self.back
            self.front.reverse()
        return self.front.pop()

    def first(self):
        # pre: not empty
        if self.front:
            return self.front[len(self.front)-1]
        else:
            return self.back[0]

    def joinItems(self, items):
        for i in items:
            self.join(i)

    def empty(self):
        return not(self.back or self.front)

    def includes(self, item):
        return item in self.front or item in self.back

    # returns a string representation of the queue - front on the left
    def string(self):
        q = []
        for ix in range(len(self.front)-1, -1, -1):      # range(): phtt!!!!!
            q.append(str(self.front[ix]))
        for item in self.back:
            q.append(str(item))
        return q

    def __bool__(self):
        return bool(self.front) or bool(self.back)

    def __iter__(self):
        for ix in range(len(self.front)-1, -1, -1):
            yield self.front[ix]
        for item in self.back:
            yield item


class MinPQueue():

    def __init__(self):
        self.queue = []
        self._makeEmpty()

    def update(self):
        self._makeUnsorted()

    def empty(self):
        return not self.queue

    def joinItems(self, it):
        for item in it:
            self.join(item)

    def includes(self, item):
        return item in self.queue

    def _joinEmpty(self, item):
        self.queue.append(item)
        self._makeSorted()

    def _leaveEmpty(self):
        return None

    def _frontEmpty(self):
        return None

    def _joinSorted(self, item):
        if item > self._frontSorted():
            self.unSorted = True
            self._makeUnsorted()
        self.queue.append(item)

    def _leaveSorted(self):
        result = self.queue.pop()
        if not self.queue:
            self._makeEmpty()
        return result

    def _frontSorted(self):
        return self.queue[len(self.queue)-1]

    def _joinUnsorted(self, item):
        self.queue.append(item)

    def _leaveUnsorted(self):
        self.queue.sort(reverse=True)
        self._makeSorted()
        return self._leaveSorted()

    def _frontUnsorted(self):
        self.queue.sort(reverse=True)
        self._makeSorted()
        return self.queue[len(self.queue)-1]

    def _makeSorted(self):
        self.join = self._joinSorted
        self.leave = self._leaveSorted
        self.front = self._frontSorted
        self.unSorted = False

    def _makeUnsorted(self):
        self.join = self._joinUnsorted
        self.leave = self._leaveUnsorted
        self.front = self._frontUnsorted
        self.unSorted = True

    def _makeEmpty(self):
        self.join = self._joinEmpty
        self.leave = self._leaveEmpty
        self.front = self._frontEmpty
        self.unSorted = False

    def __iter__(self):
        if self.unSorted:
            self.queue.sort(reverse=True)
            self._makeSorted()
        for ix in range(len(self.queue)-1, -1, -1):      # range(): phtt!!!!!
            yield self.queue[ix]

    def __str__(self):
        if self.unSorted:
            self.queue.sort(reverse=True)
            self._makeSorted()
        q = '['
        for ix in range(len(self.queue)-1, -1, -1):
            q += str(self.queue[ix])
        q +=']'
        return q


